# coding: utf-8
import collections
import contextlib

from .roles import DEFAULT_ROLE
from ._compat import implements_to_string, text_type


@contextlib.contextmanager
def processing(step):
    """
    A context manager. If an :class:`SchemaGenerationException` occurs within
    its nested code block, it adds ``step`` to it and reraises.
    """
    try:
        yield
    except SchemaGenerationException as e:
        e.steps.appendleft(step)
        raise


class Step(object):
    """A step of the schema generation process that caused the error."""

    def __init__(self, entity, role=DEFAULT_ROLE):
        """
        :param entity: An entity being processed.
        :param str role: A current role.
        """
        #: An entity being processed.
        self.entity = entity
        #: A current role.
        self.role = role

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __repr__(self):
        return '{0}({1!r}, role={2})'.format(
            self.__class__.__name__, self.entity, self.role)


@implements_to_string
class DocumentStep(Step):
    """
    A step of processing a :class:`document <.Document>`.

    :type entity: subclass of :class:`~.Document`
    """

    def __str__(self):
        return self.entity.__name__


@implements_to_string
class FieldStep(Step):
    """
    A step of processing a :class:`field <.BaseField>`.

    :type entity: instance of :class:`~.BaseField`
    """

    def __str__(self):
        return self.entity.__class__.__name__


@implements_to_string
class AttributeStep(Step):
    """
    A step of processing an attribute of a field.

    ``entity`` is the name of an attribute
    (e.g., ``"properties"``, ``"additional_properties"``, etc.)

    :type entity: str
    """

    def __str__(self):
        return self.entity


@implements_to_string
class ItemStep(Step):
    """
    A step of processing an item of an attribute.

    ``entity`` is either a key or an index (e.g., it can be ``"created_at"``
    if the current attribute is ``properties`` of a :class:`~.DictField` or
    ``0`` if the current attribute is ``items`` of a :class:`~.ArrayField`).

    :type entity: str or int
    """

    def __str__(self):
        return repr(self.entity)


@implements_to_string
class SchemaGenerationException(Exception):
    """
    Raised when a valid JSON schema can not be generated from a JSL object.

    Examples of such situation are the following:

    * A :class:`variable <.Var>` resolves to an integer but a
      :class:`.BaseField` expected;
    * All choices of :class:`.OneOfField` are variables and all resolve to ``None``.

    Note: this error can only happen if variables are used in a document or field
    description.

    :param str message: A message.
    """

    def __init__(self, message):
        self.message = message
        """A message."""
        self.steps = collections.deque()
        """
        A deque of :class:`steps <.Step>`, ordered from the first (the least specific)
        to the last (the most specific).
        """

    def _format_steps(self):
        if not self.steps:
            return ''
        parts = []
        steps = iter(self.steps)
        parts.append(str(next(steps)))
        for step in steps:
            if isinstance(step, (DocumentStep, FieldStep)):
                parts.append(' -> {0}'.format(step))
            elif isinstance(step, AttributeStep):
                parts.append('.{0}'.format(step))
            elif isinstance(step, ItemStep):
                parts.append('[{0}]'.format(step))
        return ''.join(parts)

    def __str__(self):
        rv = text_type(self.message)
        steps = self._format_steps()
        if steps:
            rv += u'\nSteps: {0}'.format(steps)
        return rv



# coding: utf-8
import inspect

from . import registry
from .exceptions import processing, DocumentStep
from .fields import BaseField, DocumentField, DictField
from .roles import DEFAULT_ROLE, Var, Scope, all_, construct_matcher, Resolvable, Resolution
from .resolutionscope import ResolutionScope
from ._compat import iteritems, iterkeys, with_metaclass, OrderedDict, Prepareable


def _set_owner_to_document_fields(cls):
    for field in cls.walk(through_document_fields=False, visited_documents=set([cls])):
        if isinstance(field, DocumentField):
            field.owner_cls = cls


class Options(object):
    """
    A container for options.

    All the arguments are the same and work exactly as for :class:`.fields.DictField`
    except ``properties`` (since it is automatically populated with the document fields)
    and these:

    :param definition_id:
        A unique string to be used as a key for this document in the "definitions"
        schema section. If not specified, will be generated from module and class names.
    :type definition_id: str or :class:`.Resolvable`
    :param str schema_uri:
        An URI of the JSON Schema meta-schema.
    :param roles_to_propagate:
        A matcher. If it returns ``True`` for a role, it will be passed to nested
        documents.
    :type roles_to_propagate: callable, string or iterable
    """

    def __init__(self, additional_properties=False, pattern_properties=None,
                 min_properties=None, max_properties=None,
                 title=None, description=None,
                 default=None, enum=None,
                 id='', schema_uri='http://json-schema.org/draft-04/schema#',
                 definition_id=None, roles_to_propagate=None):
        self.pattern_properties = pattern_properties
        self.additional_properties = additional_properties
        self.min_properties = min_properties
        self.max_properties = max_properties
        self.title = title
        self.description = description
        self.default = default
        self.enum = enum
        self.id = id

        self.schema_uri = schema_uri
        self.definition_id = definition_id
        self.roles_to_propagate = construct_matcher(roles_to_propagate or all_)


class DocumentMeta(with_metaclass(Prepareable, type)):
    """
    A metaclass for :class:`~.Document`. It's responsible for collecting
    options, fields and scopes registering the document in the registry, making
    it the owner of nested :class:`document fields <.DocumentField>` s and so on.
    """
    options_container = Options
    """
    A class to be used by :meth:`~.DocumentMeta.create_options`.
    Must be a subclass of :class:`~.Options`.
    """

    @classmethod
    def __prepare__(mcs, name, bases):
        return OrderedDict()

    def __new__(mcs, name, bases, attrs):
        fields = mcs.collect_fields(bases, attrs)
        options_data = mcs.collect_options(bases, attrs)
        options = mcs.create_options(options_data)

        attrs['_fields'] = fields
        attrs['_options'] = options
        attrs['_field'] = DictField(
            properties=fields,
            pattern_properties=options.pattern_properties,
            additional_properties=options.additional_properties,
            min_properties=options.min_properties,
            max_properties=options.max_properties,
            title=options.title,
            description=options.description,
            enum=options.enum,
            default=options.default,
            id=options.id,
        )

        klass = type.__new__(mcs, name, bases, attrs)
        registry.put_document(klass.__name__, klass, module=klass.__module__)
        _set_owner_to_document_fields(klass)
        return klass

    @classmethod
    def collect_fields(mcs, bases, attrs):
        """
        Collects fields from the current class and its parent classes.

        :rtype: a dictionary mapping field names to fields
        """
        fields = OrderedDict()
        # fields from parent classes:
        for base in reversed(bases):
            if hasattr(base, '_fields'):
                fields.update(base._fields)

        to_be_replaced = object()

        # and from the current class:
        pre_fields = OrderedDict()
        scopes = []
        for key, value in iteritems(attrs):
            if isinstance(value, (BaseField, Resolvable)):
                pre_fields[key] = value
            elif isinstance(value, Scope):
                scopes.append(value)
                for scope_key in iterkeys(value.__fields__):
                    pre_fields[scope_key] = to_be_replaced

        for name, field in iteritems(pre_fields):
            if field is to_be_replaced:
                values = []
                for scope in scopes:
                    if name in scope.__fields__:
                        values.append((scope.__matcher__, scope.__fields__[name]))
                fields[name] = Var(values)
            else:
                fields[name] = field

        return fields

    @classmethod
    def collect_options(mcs, bases, attrs):
        """
        Collects options from the current class and its parent classes.

        :returns: a dictionary of options
        """
        options = {}
        # options from parent classes:
        for base in reversed(bases):
            if hasattr(base, '_options'):
                for key, value in inspect.getmembers(base._options):
                    if not key.startswith('_') and value is not None:
                        options[key] = value

        # options from the current class:
        if 'Options' in attrs:
            for key, value in inspect.getmembers(attrs['Options']):
                if not key.startswith('_') and value is not None:
                    # HACK HACK HACK
                    if inspect.ismethod(value) and value.im_self is None:
                        value = value.im_func
                    options[key] = value
        return options

    @classmethod
    def create_options(cls, options):
        """
        Wraps ``options`` into a container class
        (see :attr:`~.DocumentMeta.options_container`).

        :param options: a dictionary of options
        :return: an instance of :attr:`~.DocumentMeta.options_container`
        """
        return cls.options_container(**options)


class Document(with_metaclass(DocumentMeta)):
    """A document. Can be thought as a kind of :class:`.fields.DictField`, which
    properties are defined by the fields and scopes added to the document class.

    It can be tuned using special ``Options`` attribute (see :class:`.Options`
    for available settings)::

        class User(Document):
            class Options(object):
                title = 'User'
                description = 'A person who uses a computer or network service.'
            login = StringField(required=True)
    """

    @classmethod
    def is_recursive(cls, role=DEFAULT_ROLE):
        """Returns ``True`` if there is a :class:`.DocumentField`-references cycle
        that contains ``cls``.

        :param str role: A current role.
        """
        for field in cls.resolve_and_walk(through_document_fields=True,
                                          role=role, visited_documents=set([cls])):
            if isinstance(field, DocumentField):
                if field.document_cls == cls:
                    return True
        return False

    @classmethod
    def get_definition_id(cls, role=DEFAULT_ROLE):
        """Returns a unique string to be used as a key for this document
        in the ``"definitions"`` schema section.
        """
        definition_id = cls._options.definition_id
        if isinstance(definition_id, Resolvable):
            definition_id = definition_id.resolve(role).value
        return definition_id or '{0}.{1}'.format(cls.__module__, cls.__name__)

    @classmethod
    def resolve_field(cls, field, role=DEFAULT_ROLE):
        """Resolves a field with the name ``field`` using ``role``.

        :raises: :class:`AttributeError`
        """
        properties = cls._field.properties
        if field in properties:
            return properties[field].resolve(role)
        else:
            return Resolution(None, role)

    @classmethod
    def resolve_and_iter_fields(cls, role=DEFAULT_ROLE):
        """The same as :meth:`.iter_fields`, but :class:`resolvables <.Resolvable>`
        are resolved using ``role``.
        """
        return cls._field.resolve_and_iter_fields(role=role)

    @classmethod
    def resolve_and_walk(cls, role=DEFAULT_ROLE, through_document_fields=False,
                         visited_documents=frozenset()):
        """The same as :meth:`.walk`, but :class:`resolvables <.Resolvable>` are
        resolved using ``role``.
        """
        fields = cls._field.resolve_and_walk(
            role=role, through_document_fields=through_document_fields,
            visited_documents=visited_documents)
        next(fields)  # we don't want to yield _field itself
        return fields

    @classmethod
    def iter_fields(cls):
        """Iterates over the fields of the document, resolving its
        :class:`resolvables <.Resolvable>` to all possible values.
        """
        return cls._field.iter_fields()

    @classmethod
    def walk(cls, through_document_fields=False, visited_documents=frozenset()):
        """
        Iterates recursively over the fields of the document, resolving
        occurring :class:`resolvables <.Resolvable>` to their all possible values.

        Visits fields in a DFS order.

        :param bool through_document_fields:
            If ``True``, walks through nested :class:`.DocumentField` fields.
        :param set visited_documents:
            Keeps track of visited :class:`documents <.Document>` to avoid infinite
            recursion when ``through_document_field`` is ``True``.
        :returns: iterable of :class:`.BaseField`
        """
        fields = cls._field.walk(through_document_fields=through_document_fields,
                                 visited_documents=visited_documents)
        next(fields)  # we don't want to yield _field itself
        return fields

    @classmethod
    def get_schema(cls, role=DEFAULT_ROLE, ordered=False):
        """Returns a JSON schema (draft v4) of the document.

        :param str role:  A role.
        :param bool ordered:
            If ``True``, the resulting schema dictionary is ordered. Fields are
            listed in the order they are added to the class. Schema properties are
            also ordered in a sensible and consistent way, making the schema more
            human-readable.
        :raises: :class:`.SchemaGenerationException`
        :rtype: dict or OrderedDict
        """
        definitions, schema = cls.get_definitions_and_schema(
            role=role, ordered=ordered,
            res_scope=ResolutionScope(base=cls._options.id, current=cls._options.id)
        )
        rv = OrderedDict() if ordered else {}
        if cls._options.id:
            rv['id'] = cls._options.id
        if cls._options.schema_uri is not None:
            rv['$schema'] = cls._options.schema_uri
        if definitions:
            rv['definitions'] = definitions
        rv.update(schema)
        return rv

    @classmethod
    def get_definitions_and_schema(cls, role=DEFAULT_ROLE, res_scope=ResolutionScope(),
                                   ordered=False, ref_documents=None):
        """Returns a tuple of two elements.

        The second element is a JSON schema of the document, and the first is
        a dictionary that contains definitions that are referenced from the schema.

        :param str role:  A role.
        :param bool ordered:
            If ``True``, the resulting schema dictionary is ordered. Fields are
            listed in the order they are added to the class. Schema properties are
            also ordered in a sensible and consistent way, making the schema more
            human-readable.
        :param res_scope:
            The current resolution scope.
        :type res_scope: :class:`~.ResolutionScope`
        :param set ref_documents:
            If subclass of :class:`.Document` is in this set, all :class:`.DocumentField` s
            pointing to it will be resolved as a reference: ``{"$ref": "#/definitions/..."}``.
            Note: resulting definitions will not contain schema for this document.
        :raises: :class:`~.SchemaGenerationException`
        :rtype: (dict, dict or OrderedDict)
        """
        is_recursive = cls.is_recursive()

        if is_recursive:
            ref_documents = set(ref_documents) if ref_documents else set()
            ref_documents.add(cls)
            res_scope = res_scope.replace(output=res_scope.base)

        with processing(DocumentStep(cls, role=role)):
            definitions, schema = cls._field.get_definitions_and_schema(
                role=role, res_scope=res_scope, ordered=ordered, ref_documents=ref_documents)

        if is_recursive:
            definition_id = cls.get_definition_id()
            definitions[definition_id] = schema
            schema = res_scope.create_ref(definition_id)

        return definitions, schema

# Remove Document itself from registry
registry.remove_document(Document.__name__, module=Document.__module__)



# coding: utf-8
import re
import sre_constants

from ..roles import Resolvable


def validate_regex(regex):
    """
    :param str regex: A regular expression to validate.
    :raises: ValueError
    """
    try:
        re.compile(regex)
    except sre_constants.error as e:
        raise ValueError('Invalid regular expression: {0}'.format(e))


def validate(value_or_var, validator):
    if isinstance(value_or_var, Resolvable):
        for value in value_or_var.iter_possible_values():
            validator(value)
    else:
        validator(value_or_var)


# coding: utf-8
from ..roles import DEFAULT_ROLE
from ..resolutionscope import EMPTY_SCOPE
from .._compat import OrderedDict
from .base import BaseSchemaField
from .util import validate, validate_regex


__all__ = [
    'StringField', 'BooleanField', 'EmailField', 'IPv4Field', 'DateTimeField',
    'UriField', 'NumberField', 'IntField', 'NullField'
]


class BooleanField(BaseSchemaField):
    """A boolean field."""

    def get_definitions_and_schema(self, role=DEFAULT_ROLE, res_scope=EMPTY_SCOPE,
                                   ordered=False, ref_documents=None):
        id, res_scope = res_scope.alter(self.id)
        schema = (OrderedDict if ordered else dict)(type='boolean')
        schema = self._update_schema_with_common_fields(schema, id=id, role=role)
        return {}, schema


class StringField(BaseSchemaField):
    """A string field.

    :param pattern:
        A regular expression (ECMA 262) that a string value must match.
    :type pattern: string or :class:`.Resolvable`
    :param format:
        A semantic format of the string (for example, ``"date-time"``,
        ``"email"``, or ``"uri"``).
    :type format: string or :class:`.Resolvable`
    :param min_length:
        A minimum length.
    :type min_length: int or :class:`.Resolvable`
    :param max_length:
        A maximum length.
    :type max_length: int or :class:`.Resolvable`
    """
    _FORMAT = None

    def __init__(self, pattern=None, format=None, min_length=None, max_length=None, **kwargs):
        if pattern is not None:
            validate(pattern, validate_regex)
        self.pattern = pattern  #:
        self.format = format or self._FORMAT  #:
        self.min_length = min_length  #:
        self.max_length = max_length  #:
        super(StringField, self).__init__(**kwargs)

    def get_definitions_and_schema(self, role=DEFAULT_ROLE, res_scope=EMPTY_SCOPE,
                                   ordered=False, ref_documents=None):
        id, res_scope = res_scope.alter(self.id)
        schema = (OrderedDict if ordered else dict)(type='string')
        schema = self._update_schema_with_common_fields(schema, id=id, role=role)

        pattern = self.resolve_attr('pattern', role).value
        if pattern:
            schema['pattern'] = pattern
        min_length = self.resolve_attr('min_length', role).value
        if min_length is not None:
            schema['minLength'] = min_length
        max_length = self.resolve_attr('max_length', role).value
        if max_length is not None:
            schema['maxLength'] = max_length
        format = self.resolve_attr('format', role).value
        if format is not None:
            schema['format'] = format
        return {}, schema


class EmailField(StringField):
    """An email field."""
    _FORMAT = 'email'


class IPv4Field(StringField):
    """An IPv4 field."""
    _FORMAT = 'ipv4'


class DateTimeField(StringField):
    """An ISO 8601 formatted date-time field."""
    _FORMAT = 'date-time'


class UriField(StringField):
    """A URI field."""
    _FORMAT = 'uri'


class NumberField(BaseSchemaField):
    """A number field.

    :param multiple_of:
        A value must be a multiple of this factor.
    :type multiple_of: number or :class:`.Resolvable`
    :param minimum:
        A minimum allowed value.
    :type minimum: number or :class:`.Resolvable`
    :param exclusive_minimum:
        Whether a value is allowed to exactly equal the minimum.
    :type exclusive_minimum: bool or :class:`.Resolvable`
    :param maximum:
        A maximum allowed value.
    :type maximum: number or :class:`.Resolvable`
    :param exclusive_maximum:
        Whether a value is allowed to exactly equal the maximum.
    :type exclusive_maximum: bool or :class:`.Resolvable`
    """
    _NUMBER_TYPE = 'number'

    def __init__(self, multiple_of=None, minimum=None, maximum=None,
                 exclusive_minimum=None, exclusive_maximum=None, **kwargs):
        self.multiple_of = multiple_of  #:
        self.minimum = minimum  #:
        self.exclusive_minimum = exclusive_minimum  #:
        self.maximum = maximum  #:
        self.exclusive_maximum = exclusive_maximum  #:
        super(NumberField, self).__init__(**kwargs)

    def get_definitions_and_schema(self, role=DEFAULT_ROLE, res_scope=EMPTY_SCOPE,
                                   ordered=False, ref_documents=None):
        id, res_scope = res_scope.alter(self.id)
        schema = (OrderedDict if ordered else dict)(type=self._NUMBER_TYPE)
        schema = self._update_schema_with_common_fields(schema, id=id, role=role)
        multiple_of = self.resolve_attr('multiple_of', role).value
        if multiple_of is not None:
            schema['multipleOf'] = multiple_of
        minimum = self.resolve_attr('minimum', role).value
        if minimum is not None:
            schema['minimum'] = minimum
        exclusive_minimum = self.resolve_attr('exclusive_minimum', role).value
        if exclusive_minimum is not None:
            schema['exclusiveMinimum'] = exclusive_minimum
        maximum = self.resolve_attr('maximum', role).value
        if maximum is not None:
            schema['maximum'] = maximum
        exclusive_maximum = self.resolve_attr('exclusive_maximum', role).value
        if exclusive_maximum is not None:
            schema['exclusiveMaximum'] = exclusive_maximum
        return {}, schema


class IntField(NumberField):
    """An integer field."""
    _NUMBER_TYPE = 'integer'


class NullField(BaseSchemaField):
    """A null field."""

    def get_definitions_and_schema(self, role=DEFAULT_ROLE, res_scope=EMPTY_SCOPE,
                                   ordered=False, ref_documents=None):
        id, res_scope = res_scope.alter(self.id)
        schema = (OrderedDict if ordered else dict)(type='null')
        schema = self._update_schema_with_common_fields(schema, id=id, role=role)
        return {}, schema



# coding: utf-8
from ..resolutionscope import ResolutionScope
from ..roles import Resolvable, Resolution, DEFAULT_ROLE


__all__ = ['BaseField', 'BaseSchemaField']


class BaseField(Resolvable):
    """A base class for fields of :class:`documents <.Document>`.
    Instances of this class may be added to a document to define its properties.

    Implements the :class:`.Resolvable` interface.

    :param required:
        If the field is required. Defaults to ``False``.
    :type required: bool or :class:`.Resolvable`
    """

    def __init__(self, required=False):
        #: Whether the field is required.
        self.required = required

    def resolve(self, role):
        """
        Implements the :class:`.Resolvable` interface.

        Always returns a ``Resolution(self, role)``.

        :rtype: :class:`.Resolution`
        """
        return Resolution(self, role)

    def iter_possible_values(self):
        """Implements the :class:`.Resolvable` interface.

        Yields a single value -- ``self``.
        """
        yield self

    def get_definitions_and_schema(self, role=DEFAULT_ROLE, res_scope=ResolutionScope(),
                                   ordered=False, ref_documents=None):  # pragma: no cover
        """Returns a tuple of two elements.

        The second element is a JSON schema of the data described by this field,
        and the first is a dictionary that contains definitions that are referenced
        from the schema.

        :param str role: A role.
        :param bool ordered:
            If ``True``, the resulting schema dictionary is ordered. Fields are
            listed in the order they are added to the class. Schema properties are
            also ordered in a sensible and consistent way, making the schema more
            human-readable.
        :param res_scope:
            The current resolution scope.
        :type res_scope: :class:`~.ResolutionScope`
        :param set ref_documents:
            If subclass of :class:`Document` is in this set, all :class:`DocumentField` s
            pointing to it will be resolved to a reference: ``{"$ref": "#/definitions/..."}``.
            Note: resulting definitions will not contain schema for this document.
        :raises: :class:`.SchemaGenerationException`
        :rtype: (dict, dict or OrderedDict)
        """
        raise NotImplementedError()

    def get_schema(self, ordered=False, role=DEFAULT_ROLE):
        """Returns a JSON schema (draft v4) of the field.

        :param str role:  A role.
        :param bool ordered:
            If ``True``, the resulting schema dictionary is ordered. Fields are
            listed in the order they are added to the class. Schema properties are
            also ordered in a sensible and consistent way, making the schema more
            human-readable.
        :raises: :class:`.SchemaGenerationException`
        :rtype: dict or OrderedDict
        """
        definitions, schema = self.get_definitions_and_schema(ordered=ordered, role=role)
        if definitions:
            schema['definitions'] = definitions
        return schema

    def resolve_attr(self, attr, role=DEFAULT_ROLE):
        """
        Resolves an attribure with the name ``field`` using ``role``.

        If the value of ``attr`` is :class:`resolvable <.Resolvable>`,
        it resolves it using a given ``role`` and returns the result.
        Otherwise it returns the raw value and ``role`` unchanged.

        :raises: :class:`AttributeError`
        :rtype: :class:`.Resolution`
        """
        value = getattr(self, attr)
        if isinstance(value, Resolvable):
            return value.resolve(role)
        return Resolution(value, role)


class BaseSchemaField(BaseField):
    """A base class for fields that directly map to JSON Schema validator.

    :param required:
        If the field is required. Defaults to ``False``.
    :type required: bool or :class:`.Resolvable`
    :param str id:
        A string to be used as a value of the `"id" keyword`_ of the resulting schema.
    :param default:
        The default value for this field. May be a callable.
    :type default: any JSON-representable object, a callable or a :class:`.Resolvable`
    :param enum:
        A list of valid choices. May be a callable.
    :type enum: list, tuple, set, callable or :class:`.Resolvable`
    :param title:
        A short explanation about the purpose of the data described by this field.
    :type title: str or :class:`.Resolvable`
    :param description:
        A detailed explanation about the purpose of the data described by this field.
    :type description: str or :class:`.Resolvable`

    .. _"id" keyword: https://tools.ietf.org/html/draft-zyp-json-schema-04#section-7.2
    """

    def __init__(self, id='', default=None, enum=None, title=None, description=None, **kwargs):
        #: A string to be used as a value of the `"id" keyword`_ of the resulting schema.
        self.id = id
        #: A short explanation about the purpose of the data.
        self.title = title
        #: A detailed explanation about the purpose of the data.
        self.description = description
        self._enum = enum
        self._default = default
        super(BaseSchemaField, self).__init__(**kwargs)

    def get_enum(self, role=DEFAULT_ROLE):
        """Returns a list to be used as a value of the ``"enum"`` schema keyword."""
        enum = self.resolve_attr('_enum', role).value
        if callable(enum):
            enum = enum()
        return enum

    def get_default(self, role=DEFAULT_ROLE):
        """Returns a value of the ``"default"`` schema keyword."""
        default = self.resolve_attr('_default', role).value
        if callable(default):
            default = default()
        return default

    def get_definitions_and_schema(self, role=DEFAULT_ROLE, res_scope=ResolutionScope(),
                                   ordered=False, ref_documents=None):  # pragma: no cover
        raise NotImplementedError()

    def _update_schema_with_common_fields(self, schema, id='', role=DEFAULT_ROLE):
        if id:
            schema['id'] = id
        title = self.resolve_attr('title', role).value
        if title is not None:
            schema['title'] = title
        description = self.resolve_attr('description', role).value
        if description is not None:
            schema['description'] = description
        enum = self.get_enum(role=role)
        if enum:
            schema['enum'] = list(enum)
        default = self.get_default(role=role)
        if default is not None:
            schema['default'] = default
        return schema

    def iter_fields(self):
        """Iterates over the nested fields of the document examining all
        possible values of the occuring :class:`resolvables <.Resolvable>`.
        """
        return iter([])

    def walk(self, through_document_fields=False, visited_documents=frozenset()):
        """Iterates recursively over the nested fields, examining all
        possible values of the occuring :class:`resolvables <.Resolvable>`.

        Visits fields in a DFS order.

        :param bool through_document_fields:
            If ``True``, walks through nested :class:`.DocumentField` fields.
        :param set visited_documents:
            Keeps track of visited :class:`documents <.Document>` to avoid infinite
            recursion when ``through_document_field`` is ``True``.
        :returns: iterable of :class:`.BaseField`
        """
        yield self
        for field in self.iter_fields():
            for field_ in field.walk(through_document_fields=through_document_fields,
                                     visited_documents=visited_documents):
                yield field_

    def resolve_and_iter_fields(self, role=DEFAULT_ROLE):
        """The same as :meth:`.iter_fields`, but :class:`resolvables <.Resolvable>`
        are resolved using ``role``.
        """
        return iter([])

    def resolve_and_walk(self, role=DEFAULT_ROLE, through_document_fields=False,
                         visited_documents=frozenset()):
        """The same as :meth:`.walk`, but :class:`resolvables <.Resolvable>` are
        resolved using ``role``.
        """
        yield self
        for field in self.resolve_and_iter_fields(role=role):
            field, field_role = field.resolve(role)
            for field_ in field.resolve_and_walk(role=field_role,
                                                 through_document_fields=through_document_fields,
                                                 visited_documents=visited_documents):
                yield field_



import sys
import inspect
from functools import wraps

from . import IS_PY3


class Prepareable(type):
    # this code is taken from https://gist.github.com/DasIch/5562625 with minor fixes
    if not IS_PY3:
        def __new__(cls, name, bases, attributes):
            try:
                constructor = attributes["__new__"]
            except KeyError:
                return type.__new__(cls, name, bases, attributes)

            def preparing_constructor(cls, name, bases, attributes):
                try:
                    cls.__prepare__
                except AttributeError:
                    return constructor(cls, name, bases, attributes)
                namespace = cls.__prepare__(name, bases)
                defining_frame = sys._getframe(1)
                for constant in reversed(defining_frame.f_code.co_consts):
                    if inspect.iscode(constant) and constant.co_name == name:
                        def get_index(attribute_name, _names=constant.co_names):
                            try:
                                return _names.index(attribute_name)
                            except ValueError:
                                return 0
                        break
                else:
                    return constructor(cls, name, bases, attributes)

                by_appearance = sorted(
                    attributes.items(), key=lambda item: get_index(item[0])
                )
                for key, value in by_appearance:
                    namespace[key] = value
                return constructor(cls, name, bases, namespace)
            attributes["__new__"] = wraps(constructor)(preparing_constructor)
            return type.__new__(cls, name, bases, attributes)


# coding: utf-8
import itertools

from .. import registry
from ..roles import DEFAULT_ROLE, Resolvable
from ..resolutionscope import EMPTY_SCOPE
from ..exceptions import SchemaGenerationException, processing, FieldStep, AttributeStep, ItemStep
from .._compat import iteritems, iterkeys, itervalues, string_types, OrderedDict
from .base import BaseSchemaField, BaseField
from .util import validate_regex


__all__ = [
    'ArrayField', 'DictField', 'OneOfField', 'AnyOfField', 'AllOfField',
    'NotField', 'DocumentField', 'RECURSIVE_REFERENCE_CONSTANT'
]

RECURSIVE_REFERENCE_CONSTANT = 'self'


class ArrayField(BaseSchemaField):
    """An array field.

    :param items:
        Either of the following:

        * :class:`BaseField` -- all items of the array must match the field schema;
        * a list or a tuple of :class:`fields <.BaseField>` -- all items of the array must be
          valid according to the field schema at the corresponding index (tuple typing);
        * a :class:`.Resolvable` resolving to either of the first two options.

    :param min_items:
        A minimum length of an array.
    :type min_items: int or :class:`.Resolvable`
    :param max_items:
        A maximum length of an array.
    :type max_items: int or :class:`.Resolvable`
    :param unique_items:
        Whether all the values in the array must be distinct.
    :type unique_items: bool or :class:`.Resolvable`
    :param additional_items:
        If the value of ``items`` is a list or a tuple, and the array length is larger than
        the number of fields in ``items``, then the additional items are described
        by the :class:`.BaseField` passed using this argument.
    :type additional_items: bool or :class:`.BaseField` or :class:`.Resolvable`
    """

    def __init__(self, items=None, additional_items=None,
                 min_items=None, max_items=None, unique_items=None, **kwargs):
        self.items = items  #:
        self.min_items = min_items  #:
        self.max_items = max_items  #:
        self.unique_items = unique_items  #:
        self.additional_items = additional_items  #:
        super(ArrayField, self).__init__(**kwargs)

    def get_definitions_and_schema(self, role=DEFAULT_ROLE, res_scope=EMPTY_SCOPE,
                                   ordered=False, ref_documents=None):
        with processing(FieldStep(self, role=role)):
            return self._do_get_definitions_and_schema(
                role=role, res_scope=res_scope, ordered=ordered, ref_documents=ref_documents)

    def _do_get_definitions_and_schema(self, role=DEFAULT_ROLE, res_scope=EMPTY_SCOPE,
                                       ordered=False, ref_documents=None):
        id, res_scope = res_scope.alter(self.id)
        schema = (OrderedDict if ordered else dict)(type='array')
        schema = self._update_schema_with_common_fields(schema, id=id, role=role)
        nested_definitions = {}

        items, items_role = self.resolve_attr('items', role)
        if items is not None:
            with processing(AttributeStep('items', role=role)):
                if isinstance(items, (list, tuple)):
                    items_schema = []
                    for i, item in enumerate(items):
                        with processing(ItemStep(i, role=items_role)):
                            if not isinstance(item, Resolvable):
                                raise SchemaGenerationException(u'{0} is not resolvable'.format(item))
                            item, item_role = item.resolve(items_role)
                            if item is None:
                                continue
                            item_definitions, item_schema = item.get_definitions_and_schema(
                                role=item_role, res_scope=res_scope,
                                ordered=ordered, ref_documents=ref_documents)
                            nested_definitions.update(item_definitions)
                            items_schema.append(item_schema)
                    if not items_schema:
                        raise SchemaGenerationException(u'Items tuple is empty')
                elif isinstance(items, BaseField):
                    items_definitions, items_schema = items.get_definitions_and_schema(
                        role=items_role, res_scope=res_scope, ordered=ordered,
                        ref_documents=ref_documents)
                    nested_definitions.update(items_definitions)
                else:
                    raise SchemaGenerationException(
                        u'{0} is not a BaseField, a list or a tuple'.format(items))
                schema['items'] = items_schema

        additional_items, additional_items_role = self.resolve_attr('additional_items', role)
        if additional_items is not None:
            with processing(AttributeStep('additional_items', role=role)):
                if isinstance(additional_items, bool):
                    schema['additionalItems'] = additional_items
                elif isinstance(additional_items, BaseField):
                    items_definitions, items_schema = additional_items.get_definitions_and_schema(
                        role=additional_items_role, res_scope=res_scope,
                        ordered=ordered, ref_documents=ref_documents)
                    schema['additionalItems'] = items_schema
                    nested_definitions.update(items_definitions)
                else:
                    raise SchemaGenerationException(
                        u'{0} is not a BaseField or a boolean'.format(additional_items))

        min_items = self.resolve_attr('min_items', role).value
        if min_items is not None:
            schema['minItems'] = min_items
        max_items = self.resolve_attr('max_items', role).value
        if max_items is not None:
            schema['maxItems'] = max_items
        unique_items = self.resolve_attr('unique_items', role).value
        if unique_items is not None:
            schema['uniqueItems'] = unique_items
        return nested_definitions, schema

    def iter_fields(self):
        rv = []
        if isinstance(self.items, (list, tuple)):
            for item in self.items:
                if isinstance(item, Resolvable):
                    rv.append(item.iter_possible_values())
        elif isinstance(self.items, Resolvable):
            for items_value in self.items.iter_possible_values():
                if isinstance(items_value, (list, tuple)):
                    for item in items_value:
                        if isinstance(item, Resolvable):
                            rv.append(item.iter_possible_values())
                else:
                    if isinstance(items_value, Resolvable):
                        rv.append(items_value.iter_possible_values())
        if isinstance(self.additional_items, Resolvable):
            rv.append(self.additional_items.iter_possible_values())
        return itertools.chain.from_iterable(rv)

    def resolve_and_iter_fields(self, role=DEFAULT_ROLE):
        items, items_role = self.resolve_attr('items', role)
        if isinstance(items, (list, tuple)):
            for item in items:
                if isinstance(item, Resolvable):
                    item_value = item.resolve(items_role).value
                    if isinstance(item_value, BaseField):
                        yield item_value
        elif isinstance(items, Resolvable):
            yield items
        additional_items = self.resolve_attr('additional_items', role).value
        if isinstance(additional_items, BaseField):
            yield additional_items


class DictField(BaseSchemaField):
    """A dictionary field.

    :param properties:
        A dictionary containing fields.
    :type properties: dict[str -> :class:`.BaseField` or :class:`.Resolvable`]
    :param pattern_properties:
        A dictionary whose keys are regular expressions (ECMA 262).
        Properties match against these regular expressions, and for any that match,
        the property is described by the corresponding field schema.
    :type pattern_properties: dict[str -> :class:`.BaseField` or :class:`.Resolvable`]
    :param additional_properties:
        Describes properties that are not described by the ``properties`` or ``pattern_properties``.
    :type additional_properties: bool or :class:`.BaseField` or :class:`.Resolvable`
    :param min_properties:
        A minimum number of properties.
    :type min_properties: int or :class:`.Resolvable`
    :param max_properties:
        A maximum number of properties
    :type max_properties: int or :class:`.Resolvable`
    """

    def __init__(self, properties=None, pattern_properties=None, additional_properties=None,
                 min_properties=None, max_properties=None, **kwargs):
        self.properties = properties  #:
        self.pattern_properties = pattern_properties  #:
        self.additional_properties = additional_properties  #:
        self.min_properties = min_properties  #:
        self.max_properties = max_properties  #:
        super(DictField, self).__init__(**kwargs)

    def _process_properties(self, properties, res_scope, ordered=False,
                            ref_documents=None, role=DEFAULT_ROLE):
        nested_definitions = {}
        schema = OrderedDict() if ordered else {}
        required = []
        for prop, field in iteritems(properties):
            with processing(ItemStep(prop, role=role)):
                if not isinstance(field, Resolvable):
                    raise SchemaGenerationException(u'{0} is not resolvable'.format(field))
                field, field_role = field.resolve(role)
                if field is None:
                    continue
                field_definitions, field_schema = field.get_definitions_and_schema(
                    role=field_role, res_scope=res_scope,
                    ordered=ordered, ref_documents=ref_documents)
                if field.resolve_attr('required', field_role).value:
                    required.append(prop)
                schema[prop] = field_schema
                nested_definitions.update(field_definitions)
        return nested_definitions, required, schema

    def get_definitions_and_schema(self, role=DEFAULT_ROLE, res_scope=EMPTY_SCOPE,
                                   ordered=False, ref_documents=None):
        with processing(FieldStep(self, role=role)):
            return self._do_get_definitions_and_schema(
                role=role, res_scope=res_scope, ordered=ordered, ref_documents=ref_documents)

    def _do_get_definitions_and_schema(self, role=DEFAULT_ROLE, res_scope=EMPTY_SCOPE,
                                       ordered=False, ref_documents=None):
        id, res_scope = res_scope.alter(self.id)
        schema = (OrderedDict if ordered else dict)(type='object')
        schema = self._update_schema_with_common_fields(schema, id=id, role=role)
        nested_definitions = {}

        with processing(AttributeStep('properties', role=role)):
            properties, properties_role = self.resolve_attr('properties', role)
            if properties is not None:
                if not isinstance(properties, dict):
                    raise SchemaGenerationException(u'{0} is not a dict'.format(properties))
                properties_definitions, properties_required, properties_schema = \
                    self._process_properties(properties, res_scope,
                                             ordered=ordered, ref_documents=ref_documents,
                                             role=properties_role)
                schema['properties'] = properties_schema
                if properties_required:
                    schema['required'] = properties_required
                nested_definitions.update(properties_definitions)

        with processing(AttributeStep('pattern_properties', role=role)):
            pattern_properties, pattern_properties_role = \
                self.resolve_attr('pattern_properties', role)
            if pattern_properties is not None:
                if not isinstance(pattern_properties, dict):
                    raise SchemaGenerationException(u'{0} is not a dict.'.format(pattern_properties))
                for key in iterkeys(pattern_properties):
                    try:
                        validate_regex(key)
                    except ValueError as e:
                        raise SchemaGenerationException(u'Invalid regexp: {0}'.format(e))
                properties_definitions, _, properties_schema = self._process_properties(
                    pattern_properties, res_scope,
                    ordered=ordered, ref_documents=ref_documents,
                    role=pattern_properties_role)
                schema['patternProperties'] = properties_schema
                nested_definitions.update(properties_definitions)

        with processing(AttributeStep('additional_properties', role=role)):
            additional_properties, additional_properties_role = \
                self.resolve_attr('additional_properties', role)
            if additional_properties is not None:
                if isinstance(additional_properties, bool):
                    schema['additionalProperties'] = additional_properties
                elif isinstance(additional_properties, BaseField):
                    additional_properties_definitions, additional_properties_schema = \
                        additional_properties.get_definitions_and_schema(
                            role=additional_properties_role, res_scope=res_scope,
                            ordered=ordered, ref_documents=ref_documents)
                    schema['additionalProperties'] = additional_properties_schema
                    nested_definitions.update(additional_properties_definitions)
                else:
                    raise SchemaGenerationException(
                        u'{0} is not a BaseField or a boolean'.format(additional_properties))

        min_properties = self.resolve_attr('min_properties', role).value
        if min_properties is not None:
            schema['minProperties'] = min_properties
        max_properties = self.resolve_attr('max_properties', role).value
        if max_properties is not None:
            schema['maxProperties'] = max_properties

        return nested_definitions, schema

    def iter_fields(self):
        def _extract_resolvables(dict_or_resolvable):
            rv = []
            possible_dicts = []
            if isinstance(dict_or_resolvable, Resolvable):
                possible_dicts = dict_or_resolvable.iter_possible_values()
            elif isinstance(dict_or_resolvable, dict):
                possible_dicts = [dict_or_resolvable]
            for possible_dict in possible_dicts:
                rv.extend(v for v in itervalues(possible_dict) if v is not None)
            return rv

        resolvables = _extract_resolvables(self.properties)
        resolvables.extend(_extract_resolvables(self.pattern_properties))
        if isinstance(self.additional_properties, Resolvable):
            resolvables.append(self.additional_properties)
        return itertools.chain.from_iterable(r.iter_possible_values() for r in resolvables)

    def resolve_and_iter_fields(self, role=DEFAULT_ROLE):
        properties, properties_role = self.resolve_attr('properties', role)
        if properties is not None:
            for field in itervalues(properties):
                field = field.resolve(properties_role).value
                if isinstance(field, BaseField):
                    yield field
        pattern_properties, pattern_properties_role = \
            self.resolve_attr('pattern_properties', role)
        if pattern_properties is not None:
            for field in itervalues(pattern_properties):
                field = field.resolve(pattern_properties_role).value
                if isinstance(field, BaseField):
                    yield field
        additional_properties = self.resolve_attr('additional_properties', role).value
        if isinstance(additional_properties, BaseField):
            yield additional_properties


class BaseOfField(BaseSchemaField):
    _KEYWORD = None

    def __init__(self, fields, **kwargs):
        self.fields = fields  #:
        super(BaseOfField, self).__init__(**kwargs)

    def get_definitions_and_schema(self, role=DEFAULT_ROLE, res_scope=EMPTY_SCOPE,
                                   ordered=False, ref_documents=None):
        with processing(FieldStep(self, role=role)):
            return self._do_get_definitions_and_schema(
                role=role, res_scope=res_scope, ordered=ordered, ref_documents=ref_documents)

    def _do_get_definitions_and_schema(self, role=DEFAULT_ROLE, res_scope=EMPTY_SCOPE,
                                       ordered=False, ref_documents=None):
        id, res_scope = res_scope.alter(self.id)
        schema = OrderedDict() if ordered else {}
        schema = self._update_schema_with_common_fields(schema, id=id)
        nested_definitions = {}

        one_of = []
        with processing(AttributeStep('fields', role=role)):
            fields, fields_role = self.resolve_attr('fields', role)
            if not isinstance(fields, (list, tuple)):
                raise SchemaGenerationException(u'{0} is not a list or a tuple'.format(fields))
            for i, field in enumerate(fields):
                with processing(ItemStep(i, role=fields_role)):
                    if not isinstance(field, Resolvable):
                        raise SchemaGenerationException(u'{0} is not resolvable'.format(field))
                    field, field_role = field.resolve(fields_role)
                    if field is None:
                        continue
                    if not isinstance(field, BaseField):
                        raise SchemaGenerationException(u'{0} is not a BaseField.'.format(field))
                    field_definitions, field_schema = field.get_definitions_and_schema(
                        role=field_role, res_scope=res_scope,
                        ordered=ordered, ref_documents=ref_documents)
                    nested_definitions.update(field_definitions)
                    one_of.append(field_schema)
            if not one_of:
                raise SchemaGenerationException(u'Fields list is empty')
        schema[self._KEYWORD] = one_of
        return nested_definitions, schema

    def iter_fields(self):
        resolvables = []
        if isinstance(self.fields, (list, tuple)):
            resolvables.extend(self.fields)
        if isinstance(self.fields, Resolvable):
            for fields in self.fields.iter_possible_values():
                if isinstance(fields, (list, tuple)):
                    resolvables.extend(fields)
                elif isinstance(fields, Resolvable):
                    resolvables.append(fields)
        return itertools.chain.from_iterable(r.iter_possible_values() for r in resolvables)

    def resolve_and_iter_fields(self, role=DEFAULT_ROLE):
        fields, fields_role = self.resolve_attr('fields', role)
        for field in fields:
            field = field.resolve(fields_role).value
            if isinstance(field, BaseField):
                yield field


class OneOfField(BaseOfField):
    """
    :param fields: A list of fields, exactly one of which describes the data.
    :type fields: list[:class:`.BaseField` or :class:`.Resolvable`]

    .. attribute:: fields
        :annotation: = None
    """
    _KEYWORD = 'oneOf'


class AnyOfField(BaseOfField):
    """
    :param fields: A list of fields, at least one of which describes the data.
    :type fields: list[:class:`.BaseField` or :class:`.Resolvable`]

    .. attribute:: fields
        :annotation: = None
    """
    _KEYWORD = 'anyOf'


class AllOfField(BaseOfField):
    """
    :param fields: A list of fields, all of which describe the data.
    :type fields: list[:class:`.BaseField` or :class:`.Resolvable`]

    .. attribute:: fields
        :annotation: = None
    """
    _KEYWORD = 'allOf'


class NotField(BaseSchemaField):
    """
    :param field: A field to negate.
    :type field: :class:`.BaseField` or :class:`.Resolvable`
    """

    def __init__(self, field, **kwargs):
        self.field = field  #:
        super(NotField, self).__init__(**kwargs)

    def iter_fields(self):
        return self.field.iter_possible_values()

    def resolve_and_iter_fields(self, role=DEFAULT_ROLE):
        field, field_role = self.resolve_attr('field', role)
        if isinstance(field, BaseField):
            yield field

    def get_definitions_and_schema(self, role=DEFAULT_ROLE, res_scope=EMPTY_SCOPE,
                                   ordered=False, ref_documents=None):
        with processing(FieldStep(self, role=role)):
            return self._do_get_definitions_and_schema(
                role=role, res_scope=res_scope, ordered=ordered, ref_documents=ref_documents)

    def _do_get_definitions_and_schema(self, role=DEFAULT_ROLE, res_scope=EMPTY_SCOPE,
                                       ordered=False, ref_documents=None):
        id, res_scope = res_scope.alter(self.id)
        schema = OrderedDict() if ordered else {}
        schema = self._update_schema_with_common_fields(schema, id=id, role=role)
        with processing(AttributeStep('field', role=role)):
            field, field_role = self.resolve_attr('field', role)
            if not isinstance(field, BaseField):
                raise SchemaGenerationException(u'{0} is not a BaseField.'.format(field))
            field_definitions, field_schema = field.get_definitions_and_schema(
                role=field_role, res_scope=res_scope,
                ordered=ordered, ref_documents=ref_documents)
        schema['not'] = field_schema
        return field_definitions, schema


class DocumentField(BaseField):
    """A reference to a nested document.

    :param document_cls:
        A string (dot-separated path to document class, i.e. ``"app.resources.User"``),
        :data:`RECURSIVE_REFERENCE_CONSTANT` or a :class:`.Document` subclass.
    :param bool as_ref:
        If ``True``, the schema of :attr:`document_cls`` is placed into the definitions
        dictionary, and the field schema just references to it:
        ``{"$ref": "#/definitions/..."}``.
        It may make a resulting schema more readable.
    """

    def __init__(self, document_cls, as_ref=False, **kwargs):
        self._document_cls = document_cls
        #: A :class:`.Document` this field is attached to.
        self.owner_cls = None
        self.as_ref = as_ref  #:
        super(DocumentField, self).__init__(**kwargs)

    def iter_fields(self):
        return self.document_cls.iter_fields()

    def walk(self, through_document_fields=False, visited_documents=frozenset()):
        yield self
        if through_document_fields:
            document_cls = self.document_cls
            if document_cls not in visited_documents:
                visited_documents = visited_documents | set([document_cls])
                for field in document_cls.walk(
                        through_document_fields=through_document_fields,
                        visited_documents=visited_documents):
                    yield field

    def resolve_and_walk(self, role=DEFAULT_ROLE, through_document_fields=False,
                         visited_documents=frozenset()):
        yield self
        if through_document_fields:
            document_cls = self.document_cls
            new_role = DEFAULT_ROLE
            if self.owner_cls:
                if self.owner_cls._options.roles_to_propagate(role):
                    new_role = role
            else:
                new_role = role
            if document_cls not in visited_documents:
                visited_documents = visited_documents | set([document_cls])
                for field in document_cls.resolve_and_walk(
                        role=new_role,
                        through_document_fields=through_document_fields,
                        visited_documents=visited_documents):
                    yield field

    def get_definitions_and_schema(self, role=DEFAULT_ROLE, res_scope=EMPTY_SCOPE,
                                   ordered=False, ref_documents=None):
        with processing(FieldStep(self, role=role)):
            return self._do_get_definitions_and_schema(
                role=role, res_scope=res_scope, ordered=ordered, ref_documents=ref_documents)

    def _do_get_definitions_and_schema(self, role=DEFAULT_ROLE, res_scope=EMPTY_SCOPE,
                                       ordered=False, ref_documents=None):
        document_cls = self.document_cls
        definition_id = document_cls.get_definition_id(role=role)
        if ref_documents and document_cls in ref_documents:
            return {}, res_scope.create_ref(definition_id)
        else:
            new_role = DEFAULT_ROLE
            if self.owner_cls:
                if self.owner_cls._options.roles_to_propagate(role):
                    new_role = role
            else:
                new_role = role
            document_definitions, document_schema = document_cls.get_definitions_and_schema(
                role=new_role, res_scope=res_scope, ordered=ordered, ref_documents=ref_documents)
            if self.as_ref:
                document_definitions[definition_id] = document_schema
                return document_definitions, res_scope.create_ref(definition_id)
            else:
                return document_definitions, document_schema

    @property
    def document_cls(self):
        """A :class:`.Document` this field points to."""
        document_cls = self._document_cls
        if isinstance(document_cls, string_types):
            if document_cls == RECURSIVE_REFERENCE_CONSTANT:
                if self.owner_cls is None:
                    raise ValueError('owner_cls is not set')
                document_cls = self.owner_cls
            else:
                try:
                    document_cls = registry.get_document(document_cls)
                except KeyError:
                    if self.owner_cls is None:
                        raise ValueError('owner_cls is not set')
                    document_cls = registry.get_document(document_cls,
                                                         module=self.owner_cls.__module__)
        return document_cls



# coding: utf-8
import collections

from ._compat import OrderedDict, iteritems, string_types


__all__ = ['all_', 'not_', 'Var', 'Scope', 'DEFAULT_ROLE']

DEFAULT_ROLE = 'default'
"""A default role."""


def all_(role):
    """
    A matcher that always returns ``True``.

    :rtype: bool
    """
    return True


def not_(*roles):
    """
    Returns a matcher that returns ``True`` for all roles
    except those are listed as arguments.

    :rtype: callable
    """
    return lambda role: role not in roles


def construct_matcher(matcher):
    if callable(matcher):
        return matcher
    elif isinstance(matcher, string_types):
        return lambda r: r == matcher
    elif isinstance(matcher, collections.Iterable):
        choices = frozenset(matcher)
        return lambda r: r in choices
    else:
        raise ValueError(
            'Unknown matcher type {} ({!r}). Only callables, '
            'strings and iterables are supported.'.format(type(matcher), matcher)
        )


Resolution = collections.namedtuple('Resolution', ['value', 'role'])
"""
A resolution result, a :class:`~collections.namedtuple`.

.. attribute:: value

    A resolved value (the first element).

.. attribute:: role

    A role to be used for visiting nested objects (the second element).
"""


class Resolvable(object):
    """An interface that represents an object which value varies
    depending on a role.
    """

    def resolve(self, role):  # pragma: no cover
        """
        Returns a value for a given ``role``.

        :param str role: A role.
        :returns: A :class:`resolution <.Resolution>`.
        """
        raise NotImplementedError()

    def iter_possible_values(self):  # pragma: no cover
        """Iterates over all possible values except ``None`` ones."""
        raise NotImplementedError()


class Var(Resolvable):
    """
    A :class:`.Resolvable` implementation.

    :param values:
        A dictionary or a list of key-value pairs, where keys are matchers
        and values are corresponding values.

        Matchers are callables returning boolean values. Strings and
        iterables are also accepted and processed as follows:

        * A string ``s`` will be replaced with a lambda ``lambda r: r == s``;
        * An iterable ``i`` will be replaced with a lambda ``lambda r: r in i``.
    :type values: dict or list of pairs

    :param default:
        A value to return if all matchers returned ``False``.

    :param propagate:
        A matcher that determines which roles are to be propagated down
        to the nested objects. Default is :data:`all_` that matches
        all roles.
    :type propagate: callable, string or iterable
    """

    def __init__(self, values=None, default=None, propagate=all_):
        self._values = []
        if values is not None:
            values = iteritems(values) if isinstance(values, dict) else values
            for matcher, value in values:
                matcher = construct_matcher(matcher)
                self._values.append((matcher, value))
        self.default = default
        self._propagate = construct_matcher(propagate)

    @property
    def values(self):
        """A list of pairs (matcher, value)."""
        return self._values

    @property
    def propagate(self):
        """A matcher that determines which roles are to be propagated down
        to the nested objects.
        """
        return self._propagate

    def iter_possible_values(self):
        """
        Implements the :class:`.Resolvable` interface.

        Yields non-``None`` values from :attr:`values`.
        """
        return (v for _, v in self._values if v is not None)

    def resolve(self, role):
        """
        Implements the :class:`.Resolvable` interface.

        :param str role: A role.
        :returns:
            A :class:`resolution <.Resolution>`,

            which value is the first value which matcher returns ``True`` and
            the role is either a given ``role`` (if :attr:`propagate`` matcher
            returns ``True``) or :data:`.DEFAULT_ROLE` (otherwise).
        """
        for matcher, matcher_value in self._values:
            if matcher(role):
                value = matcher_value
                break
        else:
            value = self.default
        new_role = role if self._propagate(role) else DEFAULT_ROLE
        return Resolution(value, new_role)


class Scope(object):
    """
    A scope consists of a set of fields and a matcher.
    Fields can be added to a scope as attributes::

        scope = Scope('response')
        scope.name = StringField()
        scope.age = IntField()

    A scope can then be added to a :class:`~.Document`.
    During a document class construction process, fields of each of its scopes
    are added to the resulting class as :class:`variables <.Var>` which only
    resolve to fields when the matcher of the scope returns ``True``.

    If two fields with the same name are assigned to different document scopes,
    the matchers of the corresponding :class:`~.Var` will be the matchers of the
    scopes in order they were added to the class.

    :class:`.Scope` can also be used as a context manager. At the moment it
    does not do anything and only useful as a syntactic sugar -- to introduce
    an extra indentation level for the fields defined within the same scope.

    For example::

        class User(Document):
            with Scope('db_role') as db:
                db._id = StringField(required=True)
                db.version = StringField(required=True)
            with Scope('response_role') as db:
                db.version = IntField(required=True)

    Is an equivalent of::

        class User(Document):
            db._id = Var([
                ('db_role', StringField(required=True))
            ])
            db.version = Var([
                ('db_role', StringField(required=True))
                ('response_role', IntField(required=True))
            ])


    :param matcher: A matcher.
    :type matcher: callable, string or iterable

    .. attribute:: __field__

        An ordered dictionary of :class:`fields <.BaseField>`.

    .. attribute:: __matcher__

        A matcher.
    """

    def __init__(self, matcher):
        # names are chosen to avoid clashing with user field names
        super(Scope, self).__setattr__('__fields__', OrderedDict())
        super(Scope, self).__setattr__('__matcher__', matcher)

    def __getattr__(self, key):
        odict = super(Scope, self).__getattribute__('__fields__')
        if key in odict:
            return odict[key]
        return super(Scope, self).__getattribute__(key)

    def __setattr__(self, key, val):
        self.__fields__[key] = val

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass
